from testfixtures import compare

from sybil.parsers.myst.lexers import (
    FencedCodeBlockLexer, DirectiveLexer,
    DirectiveInPercentCommentLexer, DirectiveInHTMLCommentLexer
)
from sybil.region import LexedRegion
from .helpers import lex


def test_fenced_code_block():
    lexer = FencedCodeBlockLexer('py?thon')
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(36, 56, {'language': 'python', 'source': '>>> 1+1\n2\n'}),
        LexedRegion(1137, 1168, {'language': 'pthon', 'source': 'assert 1 + 1 == 2\n'}),
    ])


def test_fenced_code_block_with_mapping():
    lexer = FencedCodeBlockLexer('python', mapping={'source': 'body'})
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(36, 56, {'body': '>>> 1+1\n2\n'})
    ])


def test_myst_directives():
    lexer = DirectiveLexer(directive='[^}]+', arguments='.*')
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(110, 145, {
            'directive': 'code-block', 'arguments': 'python',
            'source': '>>> 1 + 1\n3\n',
        }),
        LexedRegion(188, 273, {
            'directive': 'directivename', 'arguments': 'arguments',
            'source': 'This is\ndirective content\n',
        }),
        LexedRegion(330, 378, {
            'directive': 'eval-rst', 'arguments': '',
            'source': '.. doctest::\n\n    >>> 1 + 1\n    4\n',
        }),
        LexedRegion(1413, 1489, {
            'directive': 'foo', 'arguments': 'bar',
            'source': 'This, too, is a directive content\n',
        }),
    ])


def test_examples_from_parsing_tests():
    lexer = DirectiveLexer(directive='code-block', arguments='python')
    compare(lex('myst-codeblock.md', lexer), expected=[
        LexedRegion(99, 151, {
            'directive': 'code-block', 'arguments': 'python',
            'source': "raise Exception('boom!')\n",
        }),
        LexedRegion(701, 748, {
            'directive': 'code-block', 'arguments': 'python',
            'source': 'define_this = 1\n',
        }),
    ])


def test_myst_directives_with_mapping():
    lexer = DirectiveLexer(directive='directivename', arguments='.*', mapping={'arguments': 'foo'})
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(188, 273, {'foo': 'arguments'}),
    ])


def test_myst_percent_comment_invisible_directive():
    lexer = DirectiveInPercentCommentLexer(
        directive='(invisible-)?code(-block)?', arguments='py.*'
    )
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(449, 504, {
            'directive': 'invisible-code-block', 'arguments': 'python',
            'source': '\nb = 5\n\n...etc...\n',
        }),
        LexedRegion(584, 652, {
            'directive': 'code-block', 'arguments': 'py',
            'source': '\nb = 6\n...etc...\n\n',
        }),
    ])


def test_myst_percent_comment_invisible_directive_mapping():
    lexer = DirectiveInPercentCommentLexer(
        directive='inv[^:]+', arguments='python', mapping={'arguments': 'language'}
    )
    compare(lex('myst-lexers.md', lexer), expected=[
        LexedRegion(449, 504, {'language': 'python'}),
    ])


def test_myst_html_comment_invisible_directive():
    lexer = DirectiveInHTMLCommentLexer(
        directive='(invisible-)?code(-block)?', arguments='py.*'
    )
    compare(lex('myst-lexers.md', lexer), show_whitespace=True, expected=[
        LexedRegion(702, 827, {
            'directive': 'invisible-code-block', 'arguments': 'python',
            'source': (
                'def foo():\n'
                '   return 42\n\n'
                'meaning_of_life = 42\n\n'
                'assert foo() == meaning_of_life()\n'
            ),
        }),
        LexedRegion(912, 1015, {
            'directive': 'code-block', 'arguments': 'python',
            'source': (
                '\n'
                'blank line above ^^\n'
                '\n'
                'blank line below:\n'
                '\n'
            ),
        }),
        LexedRegion(1244, 1347, {
            'directive': 'invisible-code', 'arguments': 'py',
            'source': (
                '\n'
                'blank line above ^^\n'
                '\n'
                'blank line below:\n'
                '\n'
            ),
        }),
    ])



import re

from testfixtures import ShouldRaise, compare

from sybil import LexedRegion
from sybil.parsers.abstract.lexers import BlockLexer, LexingException
from .helpers import lex, sample_path


def test_examples_from_parsing_tests():
    lexer = BlockLexer(start_pattern=re.compile('START'), end_pattern_template='END')
    path = sample_path('lexing-fail.txt')
    with ShouldRaise(LexingException(f"Could not match 'END' in {path}:\n'\\nEDN\\n'")):
        lex('lexing-fail.txt', lexer)


def test_repr():
    compare(
        str(LexedRegion(36, 56, {'language': 'python', 'foo': None, 'source': 'X'*1000})),
        expected="<LexedRegion start=36 end=56 "
                 "{'language': 'python', 'foo': None, 'source': 'XXXXXXXXXX...'}>"
    )



from testfixtures import compare

from sybil import Document
from sybil.parsers.rest.lexers import DirectiveLexer
from sybil.region import LexedRegion
from .helpers import lex, lex_text


def test_examples_from_parsing_tests():
    lexer = DirectiveLexer(directive='code-block', arguments='python')
    compare(lex('codeblock.txt', lexer)[:2], expected=[
        LexedRegion(23, 56, {
            'directive': 'code-block', 'arguments': 'python',
            'source': 'y += 1\n',
        }),
        LexedRegion(106, 157, {
            'directive': 'code-block', 'arguments': 'python',
            'source': "raise Exception('boom!')\n",
        }),
    ])


def test_examples_from_directive_tests():
    lexer = DirectiveLexer(directive='doctest')
    compare(lex('doctest_directive.txt', lexer), expected=[
        LexedRegion(102, 136, {
            'directive': 'doctest', 'arguments': None,
            'source': '>>> 1 + 1\n2\n',
        }),
        LexedRegion(205, 249, {
            'directive': 'doctest', 'arguments': None,
            'source': '>>> 1 + 1\nUnexpected!\n',
        }),
        LexedRegion(307, 353, {
            'directive': 'doctest', 'arguments': None,
            'source': ">>> raise Exception('boom!')",
        }),
    ])


def test_directive_nested_in_md():
    lexer = DirectiveLexer(directive='doctest')
    compare(lex('doctest_rest_nested_in_md.md', lexer), expected=[
        LexedRegion(14, 47, {
            'directive': 'doctest', 'arguments': None,
            'source': '>>> 1 + 1\n3',
        }),
    ])


def test_directive_with_single_line_body_at_end_of_string():
    text = """
        My comment
    
        .. code-block:: python
            test_my_code("Hello World")
    """
    lexer = DirectiveLexer(directive='code-block')
    start = 25
    end = 95
    compare(lex_text(text, lexer), expected=[
        LexedRegion(start, end, {
            'directive': 'code-block', 'arguments': 'python',
            'source': 'test_my_code("Hello World")',
        }),
    ])
    compare(text[start:end], show_whitespace=True, expected=(
        '        .. code-block:: python\n'
        '            test_my_code("Hello World")'
    ))


def test_directive_with_multi_line_body_at_end_of_string():
    text = """
        My comment

        .. code-block:: python
            test_my_code("Hello")
            test_my_code("World")
    """
    lexer = DirectiveLexer(directive='code-block')
    start = 21
    end = 119
    compare(lex_text(text, lexer), expected=[
        LexedRegion(start, end, {
            'directive': 'code-block', 'arguments': 'python',
            'source': 'test_my_code("Hello")\ntest_my_code("World")',
        }),
    ])
    compare(text[start:end], show_whitespace=True, expected=(
        '        .. code-block:: python\n'
        '            test_my_code("Hello")\n'
        '            test_my_code("World")'
    ))



"""
This is a module doc string.
"""  # This is a comment on how annoying 3.7 and earlier are!
from functools import partial
from pathlib import Path

from testfixtures import compare

from sybil.document import PythonDocStringDocument
from .helpers import ast_docstrings, skip_if_37_or_older


@skip_if_37_or_older()
def test_extract_docstring():
    """
    This is a function docstring.
    """
    python_source_code = Path(__file__).read_text()
    expected = list(ast_docstrings(python_source_code))
    actual = list(PythonDocStringDocument.extract_docstrings(python_source_code))
    compare(expected, actual=[text for _, _, text in actual], show_whitespace=True)
    compare(expected, actual=[python_source_code[s:e] for s, e, _ in actual], show_whitespace=True)


@skip_if_37_or_older()
def test_all_docstrings_extracted_correctly(python_file):
    problems = []
    path, source = python_file
    expected = list(ast_docstrings(source))
    actual = list(PythonDocStringDocument.extract_docstrings(source))
    check = partial(compare, prefix=str(path), raises=False, show_whitespace=True)
    for check_result in (
        check(expected=len(expected), actual=len(actual)),
        check(expected, actual=[text for _, _, text in actual]),
        check(expected, actual=[source[s:e] for s, e, _ in actual]),
    ):
        if check_result:  # pragma: no cover - Only on failures!
            problems.append(check_result)
    if problems:  # pragma: no cover - Only on failures!
        raise AssertionError('docstrings not correctly extracted:\n\n'+'\n\n'.join(problems))
